import fs from "fs";
import { IWorker } from "../Worker/IWorker";
import { CSQLDatabaseWorkers } from "./SQLDatabase";
import bigInt from "big-integer";
import { IJoinedGroup } from "../Helpers/IJoinedGroup";
import { IFaucetMail } from "./IFaucetMail";
describe('CSQLDatabaseWorkers tests', () => {
    const sampleDB: CSQLDatabaseWorkers = new CSQLDatabaseWorkers('.test.sqlite');
    const sampleWorker: Omit<IWorker, "browserFingerprint" | "canBeRefferal"> = {
        faucetMail: { faucet: { username: 'test0', password: "" }, mail: {address: ''} },
        phoneNumber: "test",
        telegramID: 123456,
        apiID: 0,
        apiHash: "",
        password: "",
        stringSession: "",
        balance: 0,
        refferal: null,
        troubled: false,
        isMailed: true
    }
    const sampleGroup: Omit<IJoinedGroup, "worker" | "id"> = {
        groupID: bigInt("100000000000000"),
        timeToLeave: new Date()
    };
    test('Loading a test database', async () => {
        await expect(sampleDB.load()).resolves.not.toThrow();
    })

    test('Adding a worker to test database', async () => {
        await expect(sampleDB.addWorker(sampleWorker)).resolves.not.toThrow();
    })
    test('Checking worker from test database', async () => {
        await expect(sampleDB.getWorkers()).resolves.toHaveLength(1);
        await expect(sampleDB.getWorkers()).resolves.toMatchObject<IWorker[]>([sampleWorker as IWorker]);

    })
    test('Updating worker', async () => {
        sampleWorker.troubled=true;
        sampleWorker.balance=101;
        await expect(sampleDB.updateWorker(sampleWorker)).resolves.not.toThrow();
        const _worker=await sampleDB.getWorkers()
        expect(_worker[0]).toHaveProperty('troubled',true);
        expect(_worker[0]).toHaveProperty('balance',101);
    })
    test('Adding test groups for test worker', async () => {
        await expect(sampleDB.addGroup(sampleWorker, sampleGroup)).resolves.not.toThrow();
        await expect(sampleDB.addGroup(sampleWorker, sampleGroup)).resolves.not.toThrow();
    })

    test('Checking added groups from test database', async () => {
        await expect(sampleDB.getGroups(sampleWorker)).resolves.toHaveLength(2);
        for (const group of await sampleDB.getGroups(sampleWorker)) {
            expect(group).toMatchObject<IJoinedGroup>({ ...sampleGroup, worker: sampleWorker.phoneNumber, id: group.id });
        }
    })
    test('Deleting groups', async () => {
        await expect(sampleDB.getGroups(sampleWorker)).resolves.toHaveLength(2);
        for (const group of await sampleDB.getGroups(sampleWorker)) {
            await expect(sampleDB.removeGroup(group)).resolves.not.toThrow();
        }
        await expect(sampleDB.getGroups(sampleWorker)).resolves.toHaveLength(0);
    })
    test('Getting free wallet', async () => {
        const freeWallet: IFaucetMail = {
            faucet: { username: 'test1', password: '' },
            mail: { address: '', password: '' }
        }
        await expect(sampleDB.getWallet([sampleWorker.faucetMail, freeWallet])).resolves.toEqual(freeWallet);
        await expect(sampleDB.getWallet([sampleWorker.faucetMail])).resolves.toBeUndefined();
    })
    test('Getting *free* refferal', async () => {
        await expect(sampleDB.getRefferal()).resolves.toEqual(sampleWorker.telegramID);
    })
    test('Deleting test worker', async () => {
        expect(sampleDB.removeWorker(sampleWorker)).resolves.not.toThrow();
        await expect(sampleDB.getWorkers()).resolves.toHaveLength(0);
    })

    afterAll(async () => {
        await sampleDB.disconnect();
        if (fs.existsSync('.test.sqlite')) {
            fs.rmSync('.test.sqlite');
        }
    })
})