import { Knex } from "knex";
import knex from "knex";
import { IRawWorker, IWorker } from "../Worker/IWorker";
import { FingerprintGenerator } from "fingerprint-generator";
import { IFaucetMail } from "./IFaucetMail";
import { settings } from "./JSONDatabase";
import { IJoinedGroup } from "../Helpers/IJoinedGroup";
import bigInt from "big-integer";
import { shuffle } from "lodash";

abstract class CSQLDatabase {
    protected readonly rawDB: Knex<any, unknown[]>;
    constructor(filename: string) {
        this.rawDB = knex({
            client: "better-sqlite3",
            connection: {
                filename: filename,
            },
            useNullAsDefault: true
        });
    }
    async load() { }
    async disconnect() {
        await this.rawDB.destroy();
    }
}

export class CSQLDatabaseWorkers extends CSQLDatabase {
    constructor(filename = 'db.sqlite') {
        super(filename);
    }

    async load() {
        if ((await this.rawDB.schema.hasTable('workers')) == false) {
            this.rawDB.schema
                .createTable("workers", (table) => {
                    table
                        .string("phoneNumber", 12)
                        .notNullable().primary().unique()

                    table
                        .integer("telegramID")
                        .notNullable()
                        .unique()

                    table
                        .bigInteger("apiID")
                        .notNullable()

                    table
                        .string("apiHash")
                        .notNullable()

                    table
                        .string("password")
                        .notNullable()

                    table
                        .jsonb("faucetMail")
                        .notNullable()

                    table
                        .string("stringSession")
                        .notNullable().unique()

                    table
                        .float("balance")
                        .notNullable().defaultTo(0.0)

                    table
                        .boolean("canBeRefferal")
                        .notNullable()

                    table
                        .jsonb("browserFingerprint")
                        .notNullable().unique()
                        .comment(
                            "I save it in json data, so as not to create another table"
                        )

                    table
                        .integer("refferal")
                        .defaultTo(null)

                    table
                        .boolean("troubled")
                        .notNullable()
                        .comment("if worker has maximum groups")
                    table
                        .boolean("isMailed")
                        .notNullable()
                }).then(() => {
                    // log 
                })
        }
        if ((await this.rawDB.schema.hasTable("completedGroupsTasks")) == false) {
            this.rawDB.schema
                .createTable("completedGroupsTasks", (table) => {
                    table
                        .increments("id")
                        .primary().unique()
                    table
                        .string("worker", 12)
                        .notNullable()
                    table
                        .string("groupID")
                        .notNullable()
                    table
                        .timestamp("timeToLeave")
                        .notNullable()
                })
                .then(() => {
                    // log
                });
        }

    }
    async getRefferal() {
        const workers = await this.getWorkers();

        for (const mainWorker of workers) {
            if (workers.filter((_worker => {
                if (_worker.phoneNumber === mainWorker.phoneNumber) {
                    return false
                } else {
                    return _worker.refferal === mainWorker.telegramID
                }
            })).length < settings.maxRefferals) {
                return mainWorker.telegramID
            }
        }
        return null
    }
    async getWallet(wallets: IFaucetMail[]): Promise<IFaucetMail | undefined> {
        const workers = await this.getWorkers();
        for (const wallet of wallets) {
            if (workers.filter((worker) => {
                return worker.faucetMail.faucet.username === wallet.faucet.username
            }).length === 0) {
                return wallet
            }
        }
    }
    ;
    async addWorker(worker: Partial<Pick<IWorker,"canBeRefferal">> & Omit<IWorker, "browserFingerprint" | "troubled" | "canBeRefferal">,) {
        const fingerprintGenerator = new FingerprintGenerator();
        const fingerprint = fingerprintGenerator.getFingerprint({
            devices: ["desktop"],
            browsers: ["chrome"],
        }).fingerprint;
        if (worker.canBeRefferal){
            await this.rawDB("workers").insert(this.convertWorker({ ...worker, browserFingerprint: fingerprint, canBeRefferal: worker.canBeRefferal,troubled: false }));
        } else {
            await this.rawDB("workers").insert(this.convertWorker({ ...worker, browserFingerprint: fingerprint, canBeRefferal: Math.random() < 0.5,troubled: false }));
        }
        
    }
    async removeWorker(worker: Pick<IWorker, "phoneNumber">) {
        await this.rawDB("workers").where('phoneNumber', worker.phoneNumber).del();
    }
    async updateWorker(worker: Required<Pick<IWorker, "phoneNumber" | "troubled">> & Partial<Pick<IWorker, "balance">>) {
        if (worker.balance) {
            await this.rawDB("workers").where('phoneNumber', worker.phoneNumber).update({ balance: worker.balance, troubled: worker.troubled });
        } else {
            await this.rawDB("workers").where('phoneNumber', worker.phoneNumber).update({ troubled: worker.troubled });
        }

    }
    async getWorkers(): Promise<IWorker[]> {
        const rawWorkers = await this.rawDB.select().from<IRawWorker>("workers");
        const result=rawWorkers.map((rawWorker: IRawWorker) => {
            return {
                ...rawWorker,
                faucetMail: JSON.parse(rawWorker.faucetMail),
                browserFingerprint: JSON.parse(rawWorker.browserFingerprint),
                troubled: Boolean(rawWorker.troubled),
                isMailed: Boolean(rawWorker.isMailed)
            }
        });
        if (settings.shuffleAccounts){
            return shuffle(result)
        } else {
            return result
        }
        
    }

    convertWorker(worker: IWorker): IRawWorker {
        return { ...worker, faucetMail: JSON.stringify(worker.faucetMail), browserFingerprint: JSON.stringify(worker.browserFingerprint), isMailed: Number(worker.isMailed), troubled: Number(worker.troubled) }
    }

    async getGroups(worker: Pick<IWorker, "phoneNumber">): Promise<IJoinedGroup[]> {
        return (await this.rawDB("completedGroupsTasks").where('worker', worker.phoneNumber)).map(group => {
            return { ...group, groupID: bigInt(group.groupID as string), timeToLeave: new Date(group.timeToLeave as number) }
        })
    }
    async addGroup(worker: Pick<IWorker, "phoneNumber">, group: Omit<IJoinedGroup, "worker" | "id">) {
        await this.rawDB("completedGroupsTasks").insert({ worker: worker.phoneNumber, groupID: group.groupID.toString(), timeToleave: group.timeToLeave.getTime() });
    }
    async removeGroup(group: Pick<IJoinedGroup, "id">) {
        await this.rawDB("completedGroupsTasks").where('id', group.id).del();
    }
}

export const db = new CSQLDatabaseWorkers();