import Docker from "dockerode";
import { sleep } from "telegram/Helpers";
import { IPorts } from "../Helpers/IPorts";
import { getIP, HttpEconnRefusedError, timeout, TorError, UnknownError, UnkownHttpCodeError } from "../Helpers/utils";
import { DockerContainerIsBroken, DockerNoSuchContainer, DockerNotPulled } from "./DockerErrors";
import { Tor } from 'tor-control-ts';
import { startsWith } from "lodash";

export const DOCKERIMAGE = 'tor_privoxy_mod'

export class DockerController {
    protected static dockerAPI = new Docker({ socketPath: "/var/run/docker.sock" });
    static containers: wrappedContainer[] = [];
    static async isPulled() {
        for (const image of await this.dockerAPI.listImages()) {
            const tags = image.RepoTags;
            if (tags !== undefined) {
                for (const repoTag of tags) {
                    if (repoTag.search(DOCKERIMAGE) != -1) {
                        return true
                    }
                }
            }
        }
        return false
    }
    static async createNewDocker(ports: IPorts, timeoutMS = 300_000): Promise<wrappedContainer> {
        if (await this.isPulled()) {
            const options: Docker.ContainerCreateOptions = {
                Image: DOCKERIMAGE,
                Env: [ 'LOCATION="US"', "TOR_NewCircuitPeriod=30000"],
                HostConfig: {
                    PortBindings: {
                        "8118/tcp": [
                            {
                                HostPort: `${ports.http}`,
                            },
                        ],
                        "9050/tcp": [
                            {
                                HostPort: `${ports.socks}`,
                            },
                        ],
                        "9051/tcp": [
                            {
                                HostPort: `${ports.control}`,
                            },
                        ],
                    },
                },
                // Healthcheck: {
                //     Interval: 2 * 1000000000,
                //     Timeout: 50 * 1000000000,
                //     StartPeriod: 50 * 1000000000,
                //     Test:['CMD_SHELL',`curl -sx localhost:8118 'https://check.torproject.org/' | grep -qm1 Congratulations`]
                // },
            }
            const rawContainer = await this.dockerAPI.createContainer(options);
            let container=new wrappedContainer(ports, rawContainer,(await rawContainer.inspect()).Name,{active:false,interval:0});
            this.containers.push(container);
            await rawContainer.start();
            await sleep(60 * 1_000);
            try{
                await timeout((async ()=>{
                    while (true){
                        try {
                            await container.getIP();
                        } catch {
                            await sleep(10_000);
                            continue
                        } 
                        break
                    }
                })(),timeoutMS,new DockerContainerIsBroken())
            } catch (err){
                if (err instanceof DockerContainerIsBroken){
                    console.log(err);
                    this.containers=this.containers.filter((_container)=>{
                        return _container.ports.control!==container.ports.control
                    })
                    await container.kill();
                    return await this.createNewDocker(ports, timeoutMS)
                } else {
                    throw new DockerContainerIsBroken()
                }
            }
            this.containers=this.containers.filter((_container)=>{
                return _container.ports.control!==container.ports.control
            })
            container=new wrappedContainer(ports, rawContainer,(await rawContainer.inspect()).Name);
            this.containers.push(container);
            return container
        } else {
            throw new DockerNotPulled()
        }

    }
    static async killContainer(container: wrappedContainer) {
        this.containers = this.containers.filter((_container) => {
            return container.ports.control != _container.ports.control
        })
        return await container.kill()
    }
    
    static async killContainers() {
        for (const container of this.containers) {

            if ((await container.kill()) == false) {
                // log what something is wrong
            }
        }
        this.containers = [];
    }
}

export class wrappedContainer {
    protected rawContainer: Docker.Container;
    readonly ports: IPorts;
    readonly name: string;
    protected status: "healthy" | "unhealthy" = 'healthy';
    protected currentIP: string = '';
    protected interval: NodeJS.Timer | undefined;
    protected tor: Tor;
    constructor(ports: IPorts, rawContainer: Docker.Container,name: string, internalCheck = { active: true, interval: 60_000 }) {
        this.rawContainer = rawContainer;
        this.ports = ports;
        if (internalCheck.active == true) {
            this.interval = setInterval(this.checkLive.bind(this), internalCheck.interval);
        }
        this.tor = new Tor({
            host: 'localhost',
            port: this.ports.control,
            password: '',
        });
        this.name=name;
    }
    async restart() {
        // need check for starting
        this.status = 'unhealthy';
        try{
            await this.rawContainer.restart();
        } catch(err){

        }
        

    }
    async changeIP() { //timeoutMS = 600_000
        await this.tor.connect();
        //this.currentIP = await this.getIP();
        await this.tor.signalNewnym();
        let ip = await this.getIP();
        // let counter = 0;// crutch
        while (this.currentIP === ip) {
            // if (counter * 1_000 > timeoutMS) {
            //     this.status = 'unhealthy';
            //     throw new TorError()
            // }
            await sleep(1_000);
            ip = await this.getIP();
            // counter += 1;
        }
        this.currentIP = ip;
    }

    async getIP(maxTries = 40): Promise<string> { // this gets ip from internet
        let counter = 0;
        while (true) {
            try {
                return await getIP({
                    host: 'localhost',
                    port: this.ports.http,
                })
            } catch (err) {
                if (err instanceof TorError) {
                    if (counter < maxTries) {
                        counter += 1;
                        await sleep(10_000);
                        continue
                    } else {
                        await this.restart();
                        throw err // this means to push back worker to a list until container restarting
                    }
                } else if (err instanceof UnkownHttpCodeError) {
                    await this.restart();
                    throw err // this means to push back worker to a list until container restarting
                } else if (err instanceof HttpEconnRefusedError) {
                    try {
                        (await this.rawContainer.inspect())
                    } catch (err) {
                        this.status = 'unhealthy';
                        if (startsWith(err.reason, 'no such container')) {
                            throw new DockerNoSuchContainer()
                        }
                    }
                } else if (err instanceof UnknownError) {
                    this.status = 'unhealthy';
                    throw err
                }
                this.status = 'unhealthy';
                throw new UnknownError(err.message)
            }
        }


    }
    async checkLive() {
        try {
            (await this.rawContainer.inspect());
            this.status = 'healthy';
        } catch (err) {
            this.restart();
            this.status = 'unhealthy';
            return
        }

        try {
            await this.getIP();
            this.status = 'healthy';
        } catch (err) {
            if (err instanceof UnknownError) {
                this.restart();
            }
            this.status = 'unhealthy';
        }
    }
    async kill() {
        clearInterval(this.interval);
        // await this.tor.quit(); // not working
        try {
            await this.rawContainer.remove({ force: true, v: true });
        } catch {
            return false
        } finally {
            return true
        }

    }
    async getCurrentIP() { 
        if (this.currentIP === '') {
            try {
                this.currentIP = await this.getIP();
                return this.currentIP
            } catch { }
        }
        return this.currentIP
    }
    getHealthStatus() {
        return this.status
    }
}

