import axios, { AxiosProxyConfig, AxiosResponse } from "axios";
import { isString } from "lodash";
import { DockerController } from "../DockerWrapper/Docker";
import { EventEmitter } from 'node:events';
import Miner from "@ulixee/miner";
import { Fingerprint } from "fingerprint-generator";
import Hero from "@ulixee/hero";
import { IWorker } from "../Worker/IWorker";
import { IPorts } from "./IPorts";
import ExecuteJsPlugin from "@ulixee/execute-js-plugin";
import { settings } from "../Databases/JSONDatabase";

process.on("beforeExit", async () => {
    await DockerController.killContainers();
    console.log("Containers are closed!");
    process.exit();
})
process.on('exit', (code) => {
    console.log('Bye bye');
    console.log('Process exit event with code: ', code);

});
process.on('SIGINT', async function () {
    await DockerController.killContainers();
    console.log("Containers are closed!");
    process.exit();
});
export class UnknownError extends Error {
}
export class UnkownHttpCodeError extends Error {
}
export class HttpEconnRefusedError extends Error {
}
export class TorError extends Error {
}

export async function getIP(proxy?: AxiosProxyConfig, timeoutMS = 60 * 1_000): Promise<string> {
    let res: AxiosResponse<any, any>;
    try {
        res = await axios.get('http://icanhazip.com/', {
            proxy: proxy,
            timeout: timeoutMS,
        });
    } catch (err) {
        if (axios.isAxiosError(err)) {
            if (err.response) {
                // All responses whose are not with status code 200
                if (err.response.status === 503) {
                    // tor doesnt loaded
                    throw new TorError()
                } else {
                    // Other codes
                    throw new UnkownHttpCodeError()
                }
            } else if (err.request) {
                // Request was done, but there is no answer received
                if (err.code === "ECONNREFUSED") {
                    // is docker running?
                    throw new HttpEconnRefusedError()
                }
            }
            throw new UnknownError(err.message)
        } else {
            throw err
        }

    }
    if (res) {
        if (isString(await res.data)) {
            const ip: string = (await res.data).trim();
            if ((/^((25[0-5]|(2[0-4]|1\d|[1-9]|)\d)(\.(?!$)|$)){4}$/.test(ip))) {
                return ip
            }
        }

    }

    throw new UnknownError('Recieved bad response while getting ip:' + await res.data)



}

export function randomInRange(a: number, b: number) {
    return Math.floor(Math.random() * (b - a + 1) + b);
}
export class TimeoutError extends Error {}
export async function timeout(func: Promise<any>,timeoutMS=20_000,errorToThrow: Error=new TimeoutError(`Timeout Error after waiting ${timeoutMS}`)) {
    return await Promise.race([
        new Promise((_, reject) => {
            setTimeout(() => {
                reject(errorToThrow);
            }, timeoutMS);
        }), 
        func
    ]);
}

export async function initHero(worker: Pick<IWorker,"phoneNumber" | "browserFingerprint">, proxy: {ip: string,socks: number}, heroMiner: Miner): Promise<Hero> {
    const hero =new Hero({
        userProfile: {
            userAgentString: worker.browserFingerprint.navigator.userAgent,
            deviceProfile: {
                deviceMemory:worker.browserFingerprint.navigator.deviceMemory,
            },
        },
        name: `${worker.phoneNumber}`,
        connectionToCore: {
            host: `ws://127.0.0.1:${await heroMiner.port}`,
        },
        showChrome: settings.logLevel==="debug" ? true : false, // to debug
        userAgent: worker.browserFingerprint.navigator.userAgent,
        viewport: worker.browserFingerprint.screen,
        locale: worker.browserFingerprint.navigator.language,
        upstreamProxyUrl: `socks5://127.0.0.1:${proxy.socks}`,
        upstreamProxyIpMask:{
            ipLookupService:"icanhazip.com",
            proxyIp:proxy.ip,
            publicIp:process.env["LOCAL_IP"]
        }
    }) as Hero;
    hero.use(ExecuteJsPlugin);
    return await hero
}