import * as telegram from "telegram";
import { EntityLike } from "telegram/define";
import { isArray, isEqual, isString, isUndefined } from "lodash";
import { BadRequestError, TimedOutError } from "telegram/errors";
import { IWorker } from "./IWorker";
import { LogLevel } from "telegram/extensions/Logger";
import { settings } from "../Databases/JSONDatabase";
import { timeout, UnknownError } from "../Helpers/utils";
import { db } from "../Databases/SQLDatabase";
import { sleep, TotalList } from "telegram/Helpers";

export class NoNewMessagesError extends Error { }
export class TelegramWorker extends telegram.TelegramClient {
    readonly worker: Pick<IWorker, "phoneNumber">;
    constructor(worker: Pick<IWorker, "phoneNumber" | "apiID" | "apiHash" | "stringSession">) {
        super(
            new telegram.sessions.StringSession(worker.stringSession),
            worker.apiID,
            worker.apiHash,
            {
                deviceModel: "Samsung SM-G980",
                connectionRetries: 5,
                systemVersion: "10",
            }
        );
        this.setLogLevel(LogLevel.NONE);
        this.worker = worker;
    }
    async getMessagesByEntity(
        entity: EntityLike,
        getMessagesParams?: Omit<Partial<telegram.client.message.IterMessagesParams>, "fromUser">
    ) {
        const messages = new TotalList<telegram.Api.Message>();
        const _entity= await this.getEntity(entity);
        for (const message of await super.getMessages(_entity, getMessagesParams)) {
            if (message.sender!.id.equals(_entity.id)) {
                messages.push(message);
            }
        }
        return messages
    }
    async checkFloodErrorByBotEntity() {
        let botEntity: telegram.Api.User;
        try {
            botEntity = (await this.getEntity(
                settings.telegramLinks.botLink
            )) as telegram.Api.User;
        } catch (err) {
            if (err instanceof telegram.errors.FloodWaitError) {
                switch (err.errorMessage) {
                    case "FLOOD":
                        throw err;
                }
            }
            throw new UnknownError(`Unknown error due getEntity ${err}`);
        }
        return botEntity;
    }


    async waitForNewMessages(entity: EntityLike, idOfLastMessage: number, waitFor: number | string | string[] = 1, timeoutMS = 20 * 1_000): Promise<telegram.Api.Message[]> {
        return await timeout(
            (async ()=>{
                let messages: TotalList<telegram.Api.Message>=[];
                if (isString(waitFor)) {
                    do {
                        messages=(await this.getMessagesByEntity(entity, { minId: idOfLastMessage, limit: 100 })).filter((message) => {
                            return message.message.includes(waitFor)
                        })
                        await sleep(500);
                    } while (messages.length==0)
                    return messages
                } else if (isArray(waitFor)) {
                    do {
                        messages=(await this.getMessagesByEntity(entity, { minId: idOfLastMessage, limit: 100 })).filter((message) => {
                            return waitFor.some(_waitFor => message.message.includes(_waitFor))
                        })
                        await sleep(500);
                    } while (messages.length==0)
                    return messages
                } else {
                    do {
                        messages=(await this.getMessagesByEntity(entity, { minId: idOfLastMessage, limit: 100 }))
                        await sleep(500);
                    } while (messages.length < waitFor)
                    return messages
                }
            })(),
            timeoutMS,
            new NoNewMessagesError("No new messages")
        )
        
    }
    async isBotBanned(entity: Required<Pick<telegram.Api.User, "username">>) {
        const blockedUsers = (
            await this.invoke(
                new telegram.Api.contacts.GetBlocked({ limit: 2147483647 })
            )
        ).users as telegram.Api.User[]
        for (const blockedUser of blockedUsers) {
            if (blockedUser.username !== undefined && blockedUser.username === entity.username.replace('@', '')) {
                return true;
            }
        }
        return false

    }
    async leaveOldGroups() {
        const now = (new Date()).getTime();
        const groups = await db.getGroups(this.worker);
        // need to do
        for (const group of groups) {
            const sameGroups = groups.filter((_group) => {
                return _group.id !== group.id && _group.groupID === group.groupID
            });
            if (sameGroups.filter((_group) => { // check for oldest groups
                return _group.timeToLeave.getTime() > group.timeToLeave.getTime()
            }).length > 0) {
                await db.removeGroup(group);
                continue
            } else {
                if (group.timeToLeave.getTime() > now) {
                    let peerChannel: telegram.Api.TypeInputPeer;
                    try {
                        peerChannel = await this.getInputEntity(
                            group.groupID
                        );
                    } catch (err) {
                        console.log('Error due getInputEntity', err);
                        continue
                    }

                    try {
                        await this.invoke(
                            new telegram.Api.channels.LeaveChannel(
                                {
                                    channel: peerChannel
                                }
                            )
                        );
                    } catch (err) {
                        console.log('Error due LeaveChannel', err);
                    }

                }
            }

        }
    }
    async isJoinedToSettingsGroup() {
        for (const chat of settings.telegramLinks.groupsToJoin) {
            try {
                await this.invoke(
                    new telegram.Api.channels.GetParticipant({
                        channel: chat,
                        participant: "me",
                    })
                );
            } catch (err) {
                if (err.code == 400) {
                    try {
                        await this.invoke(
                            new telegram.Api.channels.JoinChannel({
                                channel: chat,
                            })
                        );
                    } catch (err) {
                        if (err instanceof telegram.errors.RPCError) {
                            switch (err.errorMessage) {
                                case "CHANNELS_TOO_MUCH":
                                    await db.updateWorker({
                                        phoneNumber: this.worker.phoneNumber,
                                        troubled: true
                                    });
                                    console.log(
                                        `Too much groups | ${this.worker.phoneNumber}`
                                    );
                                default:
                                    console.log(
                                        `Some error due joining HK groups | ${this.worker.phoneNumber}\n${err.message}\n${err.stack}`
                                    );

                            }
                        }
                    }
                } else {
                    console.log(
                        `Unknown error due GetParticipant | ${this.worker.phoneNumber}`
                    );
                    throw err;
                }
            }
        }
    }
    async unblockTargetBot(link: string){
        const result = await this.invoke(
            new telegram.Api.contacts.Unblock({
                id: link,
            })
        );
        if (result != true) {
            console.log(`Cant unblock ${link} | ${this.worker.phoneNumber}`);
            throw new UnknownError(`Cant unblock ${link} | ${this.worker.phoneNumber}`)
        }
    }
    async blockBot(link: string){
        return await this.invoke(
            new telegram.Api.messages.DeleteHistory({
                justClear: true,
                peer: link,
            })
        ) && await this.invoke(
            new telegram.Api.contacts.Block({
                id: link,
            })
        );
    }
    async unblockMainBot() {
        if (await this.isBotBanned(
            { username: settings.telegramLinks.botLink }
        )) {
            return await this.unblockTargetBot(settings.telegramLinks.botLink)
        }
    }
}